
import sys
import logging
import time
from datetime import datetime
import h5pyd
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Print objects in a domain in the style of the hsls utilitiy
#


cfg = Config()


#
# Usage
#
def printUsage():
    print("Usage: {} [-h] [--loglevel debug|info|warning|error] [--logfile <logfile>] [-c oonf_file] [-e endpoint] [-u username] [-p password] [-b bucket] [domain]".format(cfg["cmd"]))
    print("")
    print("Description:")
    print("    Get status information from server, or domain stats if domain is provided")
    print("")
    print("Options:")
    print("     -e | --endpoint <domain> :: The HDF Server endpoint, e.g. http://hsdshdflab.hdfgroup.org")
    print("     -u | --user <username>   :: User name credential")
    print("     -p | --password <password> :: Password credential")
    print("     -b | --bucket <bucket> :: bucket name (for use when domain is provided)")
    print("     -c | --conf <file.cnf>  :: A credential and config file")
    print("     -H | --human-readable :: with -v, print human readable sizes (e.g. 123M)")
    print("     --rescan :: refresh domain stats (for use when domain is provided)")
    print("     --logfile <logfile> :: logfile path")
    print("     --loglevel debug|info|warning|error :: Change log level")
    print("     --bucket <bucket_name> :: Storage bucket")
    print("     -h | --help    :: This message.")
    sys.exit()
#
#
#
def getUpTime(start_time):
    now = int(time.time())
    sec = now - start_time
    days = sec // (24*60*60)
    sec -= 24*60*60*days
    hrs = sec // (60*60)
    sec -= 60*60*hrs
    mins = sec // 60
    sec -= 60*mins
    if days:
        ret_str = "{} days, {} hours {} min {} sec".format(days, hrs, mins, sec)
    elif hrs:
        ret_str =  "{} hours {} min {} sec".format(hrs, mins, sec)
    elif mins:
        ret_str =  "{} min {} sec".format(mins, sec)
    else:
        ret_str =  "{} sec".format(sec)

    return ret_str

def format_size(n):
    if n is None or n == ' ':
        return ' ' * 8
    symbol = ' '
    if not cfg["human_readable"]:
        return str(n)
    # convert to common storage unit
    for s in ('B', 'K', 'M', 'G', 'T'):
        if n < 1024:
            symbol = s
            break
        n /= 1024
    if symbol == 'B':
        return "{:}B".format(n)
    else:
        return "{:.1f}{}".format(n, symbol)

def getServerInfo(cfg):
    """ get server state and print """
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    try:
        info = h5pyd.getServerInfo(username=username, password=password, endpoint=endpoint)
        print("server name: {}".format(info["name"]))
        if "state" in info:
            print("server state: {}".format(info['state']))
        print("endpoint: {}".format(endpoint))
        if "isadmin" in info and info["isadmin"]:
            admin_tag = "(admin)"
        else:
            admin_tag = ""
        print("username: {} {}".format(info["username"], admin_tag))
        print("password: {}".format(info["password"]))
        if info['state'] == "READY":
            try:
                home_folder = getHomeFolder()
                if home_folder:
                    print("home: {}".format(home_folder))
            except IOError:
                print("home: NO ACCESS",)

        if "hsds_version" in info:
            print("server version: {}".format(info["hsds_version"]))
        if "node_count" in info:
            print("node count: {}".format(info["node_count"]))
        elif "h5serv_version" in info:
            print("server version: {}".format(info["h5serv_version"]))
        if "start_time" in info:
            uptime = getUpTime(info["start_time"])
            print("up: {}".format(uptime))
        print("h5pyd version: {}".format(h5pyd.version.version))


    except IOError as ioe:
        if ioe.errno == 401:
            if username and password:
                print("username/password not valid for username: {}".format(username))
            else:
                # authentication error with openid or app token
                print("authentication failure")
        else:
            print("Error: {}".format(ioe))

def getDomainInfo(domain, cfg):
    """ get info about the domain and print """
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    if "rescan" in cfg and cfg["rescan"]:
        mode = "r+"  # need write intent
    else:
        mode = 'r'

    if domain.endswith('/'):
        is_folder = True
    else:
        is_folder = False

    try:
        if is_folder:
            f = h5pyd.Folder(domain, mode=mode, endpoint=endpoint, username=username,
                   password=password, bucket=bucket, use_cache=True)
        else:
            f = h5pyd.File(domain, mode=mode, endpoint=endpoint, username=username,
                   password=password, bucket=bucket, use_cache=False)
    except IOError as oe:
        if oe.errno in (404, 410):   # Not Found
            sys.exit("domain: {} not found".format(domain))
        elif oe.errno == 401:  # Unauthorized
            sys.exit("Authorization failure")
        elif oe.errno == 403:  # Forbidden
            sys.exit("Not allowed")
        else:
            sys.exit("Unexpected error: {}".format(oe))

    timestamp = datetime.fromtimestamp(int(f.modified))
    if not is_folder and f.last_scan:
        last_scan = datetime.fromtimestamp(int(f.last_scan))
    else:
        last_scan = None

    if is_folder:
        print("folder: {}".format(domain))
        print("    owner:           {}".format(f.owner))
        print("    last modified:   {}".format(timestamp))
    else:
        if "rescan" in cfg and cfg["rescan"]:
            f.run_scan()

        # report HDF objects (groups, datasets, and named datatypes) vs. allocated chunks
        num_objects = f.num_groups + f.num_datatypes + f.num_datasets
        if f.num_chunks > 0:
            num_chunks = f.num_chunks
        else:
            # older storeinfo format doesn't have num_chunks, so calculate
            num_chunks = f.num_objects - num_objects

        print("domain: {}".format(domain))
        print("    owner:           {}".format(f.owner))
        print("    id:              {}".format(f.id.id))
        print("    last modified:   {}".format(timestamp))
        if last_scan:
            print("    last scan:       {}".format(last_scan))
        if f.md5_sum:
            print("    md5 sum:         {}".format(f.md5_sum))
        print("    total_size:      {}".format(format_size(f.total_size)))
        print("    allocated_bytes: {}".format(format_size(f.allocated_bytes)))
        if f.metadata_bytes:
            print("    metadata_bytes:  {}".format(format_size(f.metadata_bytes)))
        if f.linked_bytes:
            print("    linked_bytes:    {}".format(format_size(f.linked_bytes)))
        print("    num objects:     {}".format(num_objects))
        print("    num chunks:      {}".format(num_chunks))
        if f.num_linked_chunks:
            print("    linked chunks:   {}".format(f.num_linked_chunks))


    f.close()

#
# Get folder in /home/ that is owned by given user
#
def getHomeFolder():
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    if not username:
        return None
    dir = h5pyd.Folder('/home/', username=username, password=password, endpoint=endpoint)  # get folder object for root
    homefolder = None
    for name in dir:
        # we should come across the given domain
        if username.startswith(name):
            # check any folders where the name matches at least part of the username
            # e.g. folder: "/home/bob/" for username "bob@acme.com"
            path = '/home/' + name + '/'
            try:
                f = h5pyd.Folder(path, username=username, password=password, endpoint=endpoint)
            except IOError as ioe:
                logging.info("find home folder - got ioe: {}".format(ioe))
                continue
            except Exception as e:
                logging.warn("find home folder - got exception: {}".format(e))
                continue
            if f.owner == username:
                homefolder = path
            f.close()
            if homefolder:
                break

    dir.close()
    return homefolder
#
# Main
#
def main():
    argn = 1
    cfg["cmd"] = sys.argv[0].split('/')[-1]
    if cfg["cmd"].endswith(".py"):
        cfg["cmd"] = "python " + cfg["cmd"]
    cfg["loglevel"] = logging.ERROR
    cfg["logfname"] = None
    cfg["human_readable"] = False
    domains = []


    while argn < len(sys.argv):
        arg = sys.argv[argn]
        val = None
        if len(sys.argv) > argn + 1:
            val = sys.argv[argn+1]
        if arg == "--loglevel":
            val = val.upper()
            if val == "DEBUG":
                cfg["loglevel"] = logging.DEBUG
            elif val == "INFO":
                cfg["loglevel"] = logging.INFO
            elif val in ("WARN", "WARNING"):
                cfg["loglevel"] = logging.WARNING
            elif val == "ERROR":
                cfg["loglevel"] = logging.ERROR
            else:
                printUsage()
            argn += 2
        elif arg == '--logfile':
            cfg["logfname"] = val
            argn += 2
        elif arg in ("-h", "--help"):
            printUsage()
        elif arg in ("-e", "--endpoint"):
            cfg["hs_endpoint"] = val
            argn += 2
        elif arg in ("-u", "--username"):
            cfg["hs_username"] = val
            argn += 2
        elif arg in ("-p", "--password"):
            cfg["hs_password"] = val
            argn += 2
        elif arg in ("-b", "--bucket"):
            cfg["hs_bucket"] = val
            argn += 2
        elif arg == "--rescan":
            cfg["rescan"] = True
            argn += 1
        elif arg == "-H":
             cfg["human_readable"] = True
             argn += 1
        else:
            domains.append(arg)
            argn += 1

    # setup logging

    logging.basicConfig(filename=cfg["logfname"], format='%(levelname)s %(asctime)s %(message)s', level=cfg["loglevel"])
    logging.debug("set log_level to {}".format(cfg["loglevel"]))

    endpoint = cfg["hs_endpoint"]
    if not endpoint or endpoint[-1] == '/' or endpoint[:4] not in ("http", "loca"):
        print("WARNING: endpoint: {} doesn't appear to be valid".format(endpoint))

    if not domains:
        getServerInfo(cfg)
    else:
        for domain in domains:
            getDomainInfo(domain, cfg)


if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

from __future__ import absolute_import

import os
import sys
import multiprocessing
try:
    from multiprocessing import shared_memory
except ImportError:
    pass  # only supported with python 3.8 or greater
import base64
import requests
import requests_unixsocket
from requests import ConnectionError
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import json
import logging

from . import openid
from .hsdsapp import HsdsApp
from .config import Config
from . import requests_lambda

MAX_CACHE_ITEM_SIZE=10000  # max size of an item to put in the cache

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


DEFAULT_TIMEOUT = 180 # seconds - allow time for hsds service to bounce


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)

class CacheResponse(object):
    """ Wrap a json response in a Requests.Response looking class.
        Note: we don't want to keep a proper requests obj in the cache since it
        would contain refernces to other objects
    """
    def __init__(self, rsp):
        # just save off what we need
        self._text = rsp.text
        self._status_code = rsp.status_code
        self._headers = rsp.headers

    @property
    def text(self):
        return self._text

    @property
    def status_code(self):
        return self._status_code

    @property
    def headers(self):
        return self._headers

def getAzureApiKey():
    """ construct API key for Active Directory if configured """
    # TBD: GoogleID?

    api_key = None

    # if Azure AD ids are set, pass them to HttpConn via api_key dict
    cfg = Config()  # pulls in state from a .hscfg file (if found).

    ad_app_id = None  # Azure AD HSDS Server id
    if "HS_AD_APP_ID" in os.environ:
        ad_app_id = os.environ["HS_AD_APP_ID"]
    elif "hs_ad_app_id" in cfg:
        ad_app_id = cfg["hs_ad_app_id"]
    ad_tenant_id = None # Azure AD tenant id
    if "HS_AD_TENANT_ID" in os.environ:
        ad_tenant_id = os.environ["HS_AD_TENANT_ID"]
    elif "hs_ad_tenant_id" in cfg:
        ad_tenant_id = cfg["hs_ad_tenant_id"]

    ad_resource_id = None # Azure AD resource id
    if "HS_AD_RESOURCE_ID" in os.environ:
        ad_resource_id = os.environ["HS_AD_RESOURCE_ID"]
    elif "hs_ad_resource_id" in cfg:
        ad_resource_id = cfg["hs_ad_resource_id"]

    ad_client_secret = None # Azure client secret
    if "HS_AD_CLIENT_SECRET" in os.environ:
        ad_client_secret = os.environ["HS_AD_CLIENT_SECRET"]
    elif "hs_ad_client_secret" in cfg:
        ad_client_secret = cfg["hs_ad_client_secret"]

    if ad_app_id and ad_tenant_id and ad_resource_id:
        # contruct dict to pass to HttpConn
        api_key = {"AD_APP_ID": ad_app_id, 
                    "AD_TENANT_ID": ad_tenant_id, 
                    "AD_RESOURCE_ID": ad_resource_id,
                    "openid_provider": "azure"}
        # optional config
        if ad_client_secret:
            api_key["AD_CLIENT_SECRET"] = ad_client_secret
    return api_key  # None if AAD not configured

def getKeycloakApiKey():
    # check for keycloak next
    cfg = Config() # pulls in state from a .hscfg file (if found).
    api_key = None
    # check to see if we are configured for keycloak authentication
    if "HS_KEYCLOAK_URI" in os.environ:
        keycloak_uri = os.environ["HS_KEYCLOAK_URI"]
    elif "hs_keycloak_uri" in cfg:
        keycloak_uri = cfg["hs_keycloak_uri"]
    else:
        keycloak_uri = None
    if "HS_KEYCLOAK_CLIENT_ID" in os.environ:
        keycloak_client_id = os.environ["HS_KEYCLOAK_CLIENT_ID"]
    elif "hs_keycloak_client_id" in cfg:
        keycloak_client_id = cfg["hs_keycloak_client_id"]
    else:
        keycloak_client_id = None
    if "HS_KEYCLOAK_REALM" in os.environ:
        keycloak_realm = cfg["HS_KEYCLOAK_REALM"]
    elif "hs_keycloak_realm" in cfg:
        keycloak_realm = cfg["hs_keycloak_realm"]
    else:
        keycloak_realm = None

    if keycloak_uri and keycloak_client_id and keycloak_uri:
        api_key = {"keycloak_uri": keycloak_uri,
                    "keycloak_client_id": keycloak_client_id,
                    "keycloak_realm": keycloak_realm,
                     "openid_provider": "keycloak" }
    return api_key
 
        
class HttpConn:
    """
    Some utility methods based on equivalents in base class.
    TBD: Should refactor these to a common base class
    """
    def __init__(self, domain_name, endpoint=None, username=None, password=None, bucket=None,
            api_key=None, mode='a', use_session=True, use_cache=True, use_shared_mem=None, logger=None, retries=3, **kwds):
        self._domain = domain_name
        self._mode = mode
        self._domain_json = None
        self._use_session = use_session
        self._retries = 1  # for testing retries
        self._hsds = None
        self._lambda = None
        self._use_shared_mem = use_shared_mem
        self._shm_block = None
        self._api_key = api_key
        self._s = None  # Sessions
        if use_cache:
            self._cache = {}
            self._objdb = {}
        else:
            self._cache = None
            self._objdb = None
        self._logger = logger
        if logger is None:
            self.log = logging
        else:
            self.log = logging.getLogger(logger)
        self.log.debug("HttpCon.init(domain: {} use_session: {} use_cache: {} retries: {})".format(domain_name, use_session, use_cache, retries))
        if endpoint is None:
            if "HS_ENDPOINT" in os.environ:
                endpoint = os.environ["HS_ENDPOINT"]
            elif "H5SERV_ENDPOINT" in os.environ:
                endpoint = os.environ["H5SERV_ENDPOINT"]

        if not endpoint:
            msg = "no endpoint set"
            raise ValueError(msg)
    
        if endpoint.startswith("lambda:"):
            # save lambda function name
            self._lambda = endpoint[len("lambda:"):]

        elif endpoint.startswith("local"):
            # create a local hsds server
            # set the number of nodes
            # if the endpoint is of the form: "local[n]", use n as the number of nodes
            # else set the number of nodes equal to number of cores
            bracket_start = endpoint.find('[')
            bracket_end = endpoint.find(']')
            dn_count = None
            if bracket_start > 0 and bracket_end > 0:
                try:
                    dn_count = int(endpoint[bracket_start+1:bracket_end])
                except ValueError:
                    # if value is '*' or something just drop down to default
                    # setup based on cpu count
                    pass 
            if not dn_count:
                dn_count = multiprocessing.cpu_count() 
                dn_count = -(-dn_count // 2)  # get the ceiling of count / 2 (don't include hyperthreading cores)
            if dn_count < 1:
                dn_count = 1
   
            hsds = HsdsApp(username=username, password=password, dn_count=dn_count, logger=self.log)
            hsds.run()
            self._hsds = hsds
            # replace 'local' with the socket path
            endpoint = hsds.endpoint     
            self.log.debug(f"got hsds endpoint: {endpoint} for 'local' connection") 

        self._endpoint = endpoint

        if username is None:
            if "HS_USERNAME" in os.environ:
                username = os.environ["HS_USERNAME"]
            elif "H5SERV_USERNAME" in os.environ:
                username = os.environ["H5SERV_USERNAME"]
        if isinstance(username, str) and (not username or username.upper() == "NONE"):
            username = None
        self._username = username

        if password is None:
            if "HS_PASSWORD" in os.environ:
                password = os.environ["HS_PASSWORD"]
            elif "H5SERV_PASSWORD" in os.environ:
                password = os.environ["H5SERV_PASSWORD"]
        if isinstance(password, str) and (not password or password.upper() == "NONE"):
            password = None
        self._password = password

        if bucket is None:
            if "HS_BUCKET" in os.environ:
                bucket = os.environ["HS_BUCKET"]
            if isinstance(bucket, str) and (not bucket or bucket.upper() == "NONE"):
                bucket = None
        self._bucket = bucket

        if api_key is None and "HS_API_KEY" in os.environ:
            api_key = os.environ["HS_API_KEY"]
        if isinstance(api_key, str) and (not api_key or api_key.upper() == "NONE"):
            api_key = None
        if not api_key:
            api_key = getAzureApiKey()
        if not api_key:
            api_key = getKeycloakApiKey()

        # Convert api_key to OpenIDHandler
        if isinstance(api_key, dict):
            # Maintain Azure-defualt backwards compatibility, but allow
            # both environment variable and kwarg override.
            # provider = Config().get('hs_openid_provider', 'azure')
            provider = api_key.get("openid_provider", "azure")
            if provider == 'azure':
                self.log.debug("creating OpenIDHandler for Azure")
                self._api_key = openid.AzureOpenID(endpoint, api_key)
            elif provider == 'google':
                self.log.debug("creating OpenIDHandler for Google")

                config = api_key.get('client_secret', None)
                scopes = api_key.get('scopes', None)
                self._api_key = openid.GoogleOpenID(endpoint, config=config, scopes=scopes)
            elif provider == 'keycloak':
                self.log.debug("creating OpenIDHandler for Keycloak")

                # for Keycloak, pass in username and password
                self._api_key = openid.KeycloakOpenID(endpoint, config=api_key, username=username, password=password)
            else:
                self.log.error("Unknown openid provider: {}".format(provider))

    def __del__(self):
        if self._hsds:
            self.log.debug('hsds stop')
            self._hsds.stop()
            self._hsds = None
        if self._s:
            self.log.debug("close session")
            self._s.close()
            self._s = None

    def getHeaders(self, username=None, password=None, headers=None):
        if headers is None:
            headers = {}
        elif "Authorization" in headers:
            return headers  # already have auth key
        if username is None:
            username = self._username
        if password is None:
            password = self._password

        if self._api_key:
            self.log.debug("using api key")
            # use OpenId handler to get a bearer token
            token = ''

            # Get a token, possibly refreshing if needed.
            if isinstance(self._api_key, openid.OpenIDHandler):
                token = self._api_key.token

            # Token was provided as a string.
            elif isinstance(self._api_key, str):
                token = self._api_key     

            if token:
                auth_string = b"Bearer " + token.encode('ascii')
                headers['Authorization'] = auth_string
        elif username is not None and password is not None:
            self.log.debug("use basic auth with username: {}".format(username))
            auth_string = username + ':' + password
            auth_string = auth_string.encode('utf-8')
            auth_string = base64.b64encode(auth_string)
            auth_string = b"Basic " + auth_string
            headers['Authorization'] = auth_string
        else:
            self.log.debug("no auth header")
            # no auth header
            pass

        return headers

    def verifyCert(self):
        # default to validate CERT for https requests, unless
        # the H5PYD_VERIFY_CERT environment variable is set and True
        #
        # TBD: set default to True once the signing authority of data.hdfgroup.org is
        # recognized
        if "H5PYD_VERIFY_CERT" in os.environ:
            verify_cert = os.environ["H5PYD_VERIFY_CERT"].upper()
            if verify_cert.startswith('F'):
                return False
        return True

    def getObjDb(self):
        return self._objdb


    def GET(self, req, format="json", params=None, headers=None, use_cache=True):
        if self._endpoint is None:
            raise IOError("object not initialized")
        # check that domain is defined (except for some specific requests)
        if req not in ("/domains", "/about", "/info", "/") and self._domain is None:
            raise IOError(f"no domain defined: req: {req}")

        if self._objdb:
            pass

        rsp = None

        headers = self.getHeaders(headers=headers)

        if params is None:
            params = {}
        if "domain" not in params:
            params["domain"] = self._domain
        if "bucket" not in params and self._bucket:
            params["bucket"] = self._bucket
        if self._api_key and not isinstance(self._api_key, dict):
            params["api_key"] = self._api_key
        self.log.debug("GET: {} [{}] bucket: {}".format(req, params["domain"], self._bucket))

        if format == "binary":
            headers['accept'] = 'application/octet-stream'

        if self._cache is not None and use_cache and format == "json" and \
            params["domain"] == self._domain and "select" not in params and \
            "query" not in params:
            self.log.debug("httpcon - checking cache")
            if req in self._cache:
                self.log.debug("httpcon - returning cache result")
                rsp = self._cache[req]
                return rsp

        self.log.info("GET: {} [{}]".format(self._endpoint + req, params["domain"]))
        for k in params:
            if k != "domain":
                v = params[k]
                self.log.debug(f"GET params {k}:{v}")
     
        try:
            if self._hsds:
                self._hsds.run()

            s = self.session
            rsp = s.get(self._endpoint + req, params=params, headers=headers, verify=self.verifyCert())
            self.log.info("status: {}".format(rsp.status_code))
            if self._hsds:
                self._hsds.run()
        except ConnectionError as ce:
            self.log.error("connection error: {}".format(ce))
            raise IOError("Connection Error")
        if rsp.status_code == 200 and self._cache is not None:
            rsp_headers = rsp.headers
            content_length = 0
            self.log.debug("conent_length: {}".format(content_length))
            if "Content-Length" in rsp_headers:
                try:
                    content_length = int(rsp_headers['Content-Length'])
                except ValueError:
                    content_length = MAX_CACHE_ITEM_SIZE + 1
            content_type = None
            if "Content-Type" in rsp_headers:
                content_type = rsp_headers['Content-Type']
            self.log.debug("content_type: {}".format(content_type))

            if content_type.startswith('application/json') and content_length < MAX_CACHE_ITEM_SIZE:
                # add to our _cache
                cache_rsp = CacheResponse(rsp)
                self.log.debug("adding {} to cache".format(req))
                self._cache[req] = cache_rsp

            if rsp.status_code == 200 and req == '/':  # and self._domain_json is None:
                self._domain_json = json.loads(rsp.text)
                self.log.info("got domain json: {}".format(self._domain_json))

        return rsp

    def PUT(self, req, body=None, format="json", params=None, headers=None):
        if self._endpoint is None:
            raise IOError("object not initialized")
        if self._domain is None:
            raise IOError("no domain defined")
        if self._cache is not None:
            # update invalidate everything in cache
            self._cache = {}
        if params:
            self.log.info("PUT params: {}".format(params))
        else:
            params = {}

        if "domain" not in params:
            params["domain"] = self._domain
        if "bucket" not in params and self._bucket:
            params["bucket"] = self._bucket
        if self._api_key:
            params["api_key"] = self._api_key

        # verify the file was open for modification
        if self._mode == 'r':
            raise IOError("Unable to create group (No write intent on file)")

        # try to do a PUT to the domain

        headers = self.getHeaders(headers=headers)

        if format == "binary":
            headers['Content-Type'] = "application/octet-stream"
            # binary write
            data = body
        else:
            data = json.dumps(body)
        self.log.info("PUT: {} format: {} [{} bytes]".format(req, format, len(data)))
       
        try:
            if self._hsds:
                self._hsds.run()
            s = self.session
            rsp = s.put(self._endpoint + req, data=data, headers=headers, params=params, verify=self.verifyCert())
            self.log.info("status: {}".format(rsp.status_code))
            if self._hsds:
                self._hsds.run()
        except ConnectionError as ce:
            self.log.error("connection error: {}".format(ce))
            raise IOError("Connection Error")

        if rsp.status_code == 201 and req == '/':
            self.log.info("clearing domain_json cache")
            self._domain_json = None
        self.log.info("PUT returning: {}".format(rsp))
        return rsp

    def POST(self, req, body=None, format="json", params=None, headers=None):
        if self._endpoint is None:
            raise IOError("object not initialized")
        if self._domain is None:
            raise IOError("no domain defined")
        if self._cache is not None:
            # invalidate cache for updates
            # TBD: handle special case for point selection since that doesn't modify anything
            self._cache = {}

        if params is None:
            params = {}
        if "domain" not in params:
            params["domain"] = self._domain
        if "bucket" not in params and self._bucket:
            params["bucket"] = self._bucket
        if self._api_key:
            params["api_key"] = self._api_key

        # verify we have write intent (unless this is a dataset point selection)
        if req.startswith("/datasets/") and req.endswith("/value"):
            point_sel = True
        else:
            point_sel = False
        if self._mode == 'r' and not point_sel:
            raise IOError("Unable perform request (No write intent on file)")

        # try to do a POST to the domain

        headers = self.getHeaders(headers=headers)

        if format=="binary":
            # For POST, binary we send and recieve data as binary
            headers['Content-Type'] = "application/octet-stream"
            headers['accept'] = 'application/octet-stream'
            # binary write
            data = body
        else:
            data = json.dumps(body)

        self.log.info("POST: " + req)
  
        try:
            s = self.session
            rsp = s.post(self._endpoint + req, data=data, headers=headers, params=params, verify=self.verifyCert())
        except ConnectionError as ce:
            self.log.warn("connection error: ", ce)
            raise IOError(str(ce))
        
        if rsp.status_code not in (200, 201):
            self.log.error("POST error: {}".format(rsp.status_code))

        return rsp

    def DELETE(self, req, params=None, headers=None):
        if self._endpoint is None:
            raise IOError("object not initialized")
        if self._cache is not None:
            self._cache = {}
        if req not in ("/domains", "/") and self._domain is None:
            raise IOError("no domain defined")
        if params is None:
            params = {}
        if "domain" not in params:
            params["domain"] = self._domain
        if "bucket" not in params and self._bucket:
            params["bucket"] = self._bucket
        if self._api_key:
            params["api_key"] = self._api_key

        # verify we have write intent
        if self._mode == 'r':
            raise IOError("Unable perform request (No write intent on file)")

        # try to do a DELETE of the resource

        headers = self.getHeaders(headers=headers)

        self.log.info("DEL: " + req)
        try:
            s = self.session
            rsp = s.delete(self._endpoint + req, headers=headers, params=params, verify=self.verifyCert())
            self.log.info("status: {}".format(rsp.status_code))
        except ConnectionError as ce:
            self.log.error("connection error: {}".format(ce))
            raise IOError("Connection Error")

        if rsp.status_code == 200 and req == '/':
            self.log.info("clearning domain_json cache")
            self._domain_json = None

        return rsp

    @property
    def session(self):
        # create a session object to re-use http connection when possible
        s = requests
        retries=self._retries
        backoff_factor=1
        status_forcelist=(500, 502, 503, 504)
        allowed_methods = ["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST"]  # include POST retries
        if self._use_session:
            if self._s is None:
                if self._endpoint.startswith("http+unix://"):
                    self.log.debug("create unixsocket session")
                    s = requests_unixsocket.Session()
                elif self._endpoint.startswith("http+lambda://"):
                    s = requests_lambda.Session()
                else:
                    # regular request session
                    s = requests.Session()

                retry = Retry(
                    total=retries,
                    read=retries,
                    connect=retries,
                    backoff_factor=backoff_factor,
                    status_forcelist=status_forcelist,
                    allowed_methods=allowed_methods
                )
             
                s.mount('http://', TimeoutHTTPAdapter(max_retries=retry))
                s.mount('https://', TimeoutHTTPAdapter(max_retries=retry))
                self._s = s
            else:
                s = self._s
        return s

    def close(self):
        if self._s:
            self._s.close()
            self._s = None
        if self._hsds:
            self._hsds.stop()
            self._hsds = None
        if self._shm_block:
            self._shm_block.close()
            self._shm_block.unlink()
            self._shm_block = None

    @property
    def domain(self):
        return self._domain

    @property
    def username(self):
        return self._username

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def password(self):
        return self._password

    @property
    def mode(self):
        return self._mode

    @property
    def cache_on(self):
        if self._cache is None:
            return False
        else:
            return True

    @property
    def use_shared_mem(self):
        if self._use_shared_mem:
            return True
        else: 
            return False
        
        # TBD: use shared mem for socket connections?
        """
        is None or not 
            if self._endpoint.startswith("http+unix"):
                return True
            else:
                return False
        else:
            return self._use_shared_mem
        """

    def get_shm_buffer(self, min_size=None):
        if not self.use_shared_mem:
            return None
        if sys.version_info.major != 3 or sys.version_info.minor < 8:
            return None  # need at least python 3.8
        if self._shm_block and (min_size is None or self._shm_block.size >= min_size):
            # just return existing shared memory block
            self.log.debug(f"returing shm_block - size: {self._shm_block.size}")
            return self._shm_block.buf
        # Free exiting block if any
        if self._shm_block:
            self._shm_block.close()
            self._shm_block.unlink()
            self._shm_block = None
        self.log.debug(f"allocating shm block - size: {min_size}")
        self._shm_block = shared_memory.SharedMemory(create=True, size=min_size)
        return self._shm_block.buf

    @property
    def shm_buffer_name(self):
        if not self._shm_block:
            return None
        else:
            return self._shm_block.name  

    @property
    def domain_json(self):
        if self._domain_json is None:
            rsp = self.GET('/')
            if rsp.status_code != 200:
                raise IOError(rsp.reason)
            # assume JSON
            self._domain_json = json.loads(rsp.text)
        return self._domain_json

    @property
    def root_uuid(self):
        domain_json = self.domain_json
        if "root" not in domain_json:
            raise IOError("Unexpected response")
        root_uuid = domain_json["root"]
        return root_uuid

    @property
    def compressors(self):
        compressors = []
        if "compressors" in self.domain_json:
            compressors = self.domain_json["compressors"]
        if not compressors:
            compressors = ["gzip",]
        return compressors

    @property
    def modified(self):
        """Last modified time of the domain as a datetime object."""
        domain_json = self.domain_json
        if "lastModified" not in domain_json:
            raise IOError("Unexpected response")
        last_modified = domain_json["lastModified"]
        return last_modified

    @property
    def created(self):
        """Creation time of the domain"""
        domain_json = self.domain_json
        if "created" not in domain_json:
            raise IOError("Unexpected response")
        created = domain_json["created"]
        return created

    @property
    def owner(self):
        """ username of creator of domain"""
        domain_json = self.domain_json
        username = None
        if 'owner' in domain_json:
            # currently this is only available for HSDS
            username = domain_json["owner"]
        return username

    @property
    def logging(self):
        """ return name of logging handler"""
        return self.log



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import sys
import os
import time
import logging
import numpy as np

USE_H5PY=0
if USE_H5PY:
    import h5py
else:
    import h5pyd as h5py

loglevel = logging.DEBUG
logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)

dims = [16, 16, 16]

if len(sys.argv) < 2 or sys.argv[1] in ("-h", "--help"):
    print("usage: python writedatacube.py <hspath> [h] [w] [d]")
    print("example: python writedatacube.py /home/joebob 16")
    sys.exit(1)

folder = sys.argv[1]

if len(sys.argv) > 2:
    dims[0] = int(sys.argv[2])
    dims[1] = dims[2] = dims[0]

if len(sys.argv) > 3:
    dims[1] = int(sys.argv[3])
    dims[2] = dims[0]

if len(sys.argv) > 4:
    dims[2] = int(sys.argv[4])

filename = os.path.join(folder, "cube")
for s in dims:
    filename += '_' + str(s)
filename += ".h5"
print("filename:", filename)


f = h5py.File(filename, "w")   

print("create dataset")

dset = f.create_dataset('dset', dims,  dtype='int8')

print("name:", dset.name)
print("shape:", dset.shape)
print("chunks:", dset.chunks)
print("dset.type:", dset.dtype)
print("dset.maxshape:", dset.maxshape)
print("bytes per slice: ", dims[0]*dims[1])
if isinstance(f.id.id, str) and f.limits:
    print("max_request_size:", f.limits["max_request_size"])

print("writing data...")

for i in range(dims[2]):
    now = time.time()
    arr = np.zeros((dims[0], dims[1]), dtype=dset.dtype)
    mid_x = dims[0] // 2
    mid_y = dims[1] // 2
    arr[mid_x,:] = 22
    arr[:,mid_y] = 44
    dset[:,:,i] = arr
    elapse = time.time() - now
    print("{:.4f}s".format(elapse))

print("closing file...")
f.close()

print("done!")



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import subprocess
import logging
from common import ut, TestCase


class TestHsinfo(TestCase):


    def test_help(self):

        arg = "-h"
        result = subprocess.check_output(["python", "../../h5pyd/_apps/hsinfo.py", "-h"])
        self.assertTrue(len(result) > 400)

if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()




import os
import sys
import signal
import subprocess
import time
import uuid
import queue
import threading
import logging


def _enqueue_output(out, queue, loglevel):
    for line in iter(out.readline, b''):
        # filter lines by loglevel
        words = line.split()
        put_line = True

        if loglevel != logging.DEBUG:
            if len(words) >= 2:
                # format should be "node_name log_level> msg"
                level = words[1][:-1]
                if loglevel == logging.INFO:
                    if level == "DEBUG":
                        put_line = False
                elif loglevel == logging.WARN or loglevel == logging.WARNING:
                    if not level.startswith("WARN") and level != "ERROR":
                        put_line = False
                elif loglevel == logging.ERROR:
                    if level != "ERROR":
                        put_line = False

        if put_line:
            queue.put(line)
    logging.debug("enqueu_output close()")
    out.close()


class HsdsApp:
    """
    Class to initiate and manage sub-process HSDS service
    """

    def __init__(self, username=None, password=None, logger=None, dn_count=1, logfile=None):
        """
        Initializer for class
        """

        """
        # Using tempdir is causing a unicode exception
        # See: https://bugs.python.org/issue32958
        self._tempdir = tempfile.TemporaryDirectory()
        tmp_dir = self._tempdir.name
        """
        # create a random dirname
        tmp_dir = "/tmp"  # TBD: will this work on windows?
        rand_name = uuid.uuid4().hex[:8]
        socket_dir = f"{tmp_dir}/hs{rand_name}/"  # TBD: use temp dir
        self._dn_urls = []
        self._socket_paths = []
        self._processes = []
        self._queues = []
        self._threads = []
        self._dn_count = dn_count
        self._username = username
        self._password = password
        self._logfile = None

        if logger is None:
            self.log = logging
        else:
            self.log = logger

        os.mkdir(socket_dir)

        self.log.debug(f"HsdsApp init - Using socketdir: {socket_dir}")

        # url-encode any slashed in the socket dir
        socket_url = ""
        for ch in socket_dir:
            if ch == '/':
                socket_url += "%2F"
            else:
                socket_url += ch

        for i in range(dn_count):
            socket_name = f"dn_{(i+1)}.sock"
            dn_url = f"http+unix://{socket_url}{socket_name}"
            self._dn_urls.append(dn_url)
            self._socket_paths.append(f"{socket_dir}{socket_name}")

        # sort the ports so that node_number can be determined based on dn_url
        self._dn_urls.sort()
        self._endpoint = f"http+unix://{socket_url}sn_1.sock"
        self._socket_paths.append(f"{socket_dir}sn_1.sock")
        self._rangeget_url = f"http+unix://{socket_url}rangeget.sock"
        self._socket_paths.append(f"{socket_dir}rangeget.sock")

    @property
    def endpoint(self):
        return self._endpoint

    def print_process_output(self):
        """ print any queue output from sub-processes
        """
        # print("print_process_output")

        while True:
            got_output = False
            for q in self._queues:
                try:
                    line = q.get_nowait()  # or q.get(timeout=.1)
                except queue.Empty:
                    pass  # no output on this queue yet
                else:
                    if isinstance(line, bytes):
                        # self.log.debug(line.decode("utf-8").strip())
                        print(line.decode("utf-8").strip())
                    else:
                        print(line.strip())
                    got_output = True
            if not got_output:
                break  # all queues empty for now

    def check_processes(self):
        self.log.debug("check processes")
        self.print_process_output()
        for p in self._processes:
            if p.poll() is not None:
                result = p.communicate()
                msg = f"process {p.args[0]} ended, result: {result}"
                self.log.warn(msg)
                # TBD - restart failed process

    def run(self):
        """ startup hsds processes
        """
        if self._processes:
            # just check process state and restart if necessary
            self.check_processes()
            return

        dn_urls_arg = ""
        for dn_url in self._dn_urls:
            if dn_urls_arg:
                dn_urls_arg += ','
            dn_urls_arg += dn_url

        pout = subprocess.PIPE   # will pipe to parent
        # create processes for count dn nodes, sn node, and rangeget node
        count = self._dn_count + 2  # plus 2 for rangeget proxy and sn
        # set PYTHONUNBUFFERED so we can get any output immediately
        os.environ["PYTHONUNBUFFERED"] = "1"
        # TODO: don't modify parent process env, use os.environ.copy(), set, and popen(env=)

        common_args = ["--standalone", ]
        # print("setting log_level to:", args.loglevel)
        # common_args.append(f"--log_level={args.loglevel}")
        common_args.append(f"--dn_urls={dn_urls_arg}")
        common_args.append(f"--rangeget_url={self._rangeget_url}")
        common_args.append(f"--hsds_endpoint={self._endpoint}")
        common_args.append("--server_name=Direct Connect (HSDS)")
        common_args.append("--password_file=")
        common_args.append("--use_socket")

        py_exe = sys.executable
        cmd_dir = os.path.join(sys.exec_prefix, "bin")
        for i in range(count):
            if i == 0:
                # args for service node
                pargs = [py_exe,
                         os.path.join(cmd_dir, "hsds-servicenode"),
                         "--log_prefix=sn "]
                if self._username:
                    pargs.append(f"--hs_username={self._username}")
                if self._password:
                    pargs.append(f"--hs_password={self._password}")
                pargs.append(f"--sn_url={self._endpoint}")
                pargs.append("--logfile=sn1.log")
            elif i == 1:
                # args for rangeget node
                pargs = [py_exe,
                         os.path.join(cmd_dir, "hsds-rangeget"),
                         "--log_prefix=rg "]
            else:
                node_number = i - 2  # start with 0
                pargs = [py_exe,
                         os.path.join(cmd_dir, "hsds-datanode"),
                         f"--log_prefix=dn{node_number+1} "]
                pargs.append(f"--dn_urls={dn_urls_arg}")
                pargs.append(f"--node_number={node_number}")
            # logging.info(f"starting {pargs[0]}")
            pargs.extend(common_args)
            p = subprocess.Popen(pargs, bufsize=1, universal_newlines=True, shell=False, stdout=pout)
            self._processes.append(p)
            if not self._logfile:
                # setup queue so we can check on process output without blocking
                q = queue.Queue()
                loglevel = self.log.root.level
                t = threading.Thread(target=_enqueue_output, args=(p.stdout, q, loglevel))
                self._queues.append(q)
                t.daemon = True  # thread dies with the program
                t.start()
                self._threads.append(t)

        # wait to sockets are initialized
        start_ts = time.time()
        SLEEP_TIME = 0.1  # time to sleep between checking on socket connection
        MAX_INIT_TIME = 10.0  # max time to wait for socket to be initialized

        while True:
            ready = 0
            for socket_path in self._socket_paths:
                if os.path.exists(socket_path):
                    ready += 1
            if ready == count:
                self.log.info("all processes ready!")
                break
            else:
                self.log.debug(f"{ready}/{count} ready")
                self.log.debug(f"sleeping for {SLEEP_TIME}")
                time.sleep(SLEEP_TIME)
                if time.time() > start_ts + MAX_INIT_TIME:
                    msg = f"failed to initialize after {MAX_INIT_TIME} seconds"
                    self.log.error(msg)
                    raise IOError(msg)

        self.log.info(f"Ready after: {(time.time()-start_ts):4.2f} s")

    def stop(self):
        """ terminate hsds processes
        """
        if not self._processes:
            return
        now = time.time()
        logging.info(f"hsds app stop at {now}")
        for p in self._processes:
            logging.info(f"sending SIGINT to {p.args[0]}")
            p.send_signal(signal.SIGINT)
        # wait for sub-proccesses to exit
        SLEEP_TIME = 0.1  # time to sleep between checking on process state
        MAX_WAIT_TIME = 10.0  # max time to wait for sub-process to terminate
        start_ts = time.time()
        while True:
            is_alive = False
            for p in self._processes:
                if p.poll() is None:
                    is_alive = True
            if is_alive:
                logging.debug(f"still alive, sleep {SLEEP_TIME}")
                time.sleep(SLEEP_TIME)
            else:
                logging.debug("all subprocesses exited")
                break
            if time.time() > start_ts + MAX_WAIT_TIME:
                msg = f"failed to terminate after {MAX_WAIT_TIME} seconds"
                self.log.error(msg)
                break

        # kill any reluctant to die processes
        for p in self._processes:
            if p.poll():
                logging.info(f"terminating {p.args[0]}")
                p.terminate()
        self._processes = []
        for t in self._threads:
            del t
        self._threads = []

    def __del__(self):
        """ cleanup class resources """
        self.stop()
        # self._tempdir.cleanup()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import logging
import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

import os.path as op
from common import ut, TestCase


class TestFolders(TestCase):


    def test_list(self):
        #loglevel = logging.DEBUG
        #logging.basicConfig( format='%(asctime)s %(message)s', level=loglevel)
        test_domain = self.getFileName("folder_test")


        filepath = self.getPathFromDomain(test_domain)
        print(filepath)
        # create test file if not present.
        # on first run, this may take a minute before it is visible as a folder item
        f = h5py.File(filepath, mode='a')
        if config.get("use_h5py"):
            # Folders not supported for h5py
            f.close()
            return

        self.assertTrue(f.id.id is not None)
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # HSDS currently supports folders, but h5serv does not
            f.close()
            return
        f.close()

        folder_name = op.dirname(filepath)  + '/'

        d = h5py.Folder(folder_name)  # get folder object    

        self.assertEqual(d.domain, folder_name)
        self.assertTrue(d.modified)
        self.assertTrue(d.created)
        self.assertEqual(str(d), folder_name)
        self.assertEqual(d.owner, self.test_user1["name"])

        dir_parent = d.parent
        self.assertEqual(dir_parent[:-1], op.dirname(folder_name[:-1]))

        # get ACL for dir
        dir_acl = d.getACL(self.test_user1["name"])
        self.assertEqual(len(dir_acl.keys()), 7)
        for k in dir_acl.keys():
            self.assertTrue(dir_acl[k])

        dir_acls = d.getACLs()
        self.assertTrue(isinstance(dir_acls, list))

        count = len(d)
        self.assertTrue(count > 1)

        test_domain_found = False

        i = 0
        for name in d:
            if name == op.basename(test_domain):
                self.assertFalse(test_domain_found)
                test_domain_found = True
            item = d[name]
            #'owner': 'test_user1',
            #'created': 1496729517.2346532,
            #'class': 'domain',
            #'name': '/org/hdfgroup/h5pyd_test/bool_dset',
            #'lastModified': 1496729517.2346532
            #self.assertTrue("created" in item)
            self.assertTrue("owner" in item)
            self.assertTrue("class" in item)
            self.assertTrue("name" in item)
            if "root" in item:
                # non-folder objects will have last modified time
                self.assertTrue("lastModified" in item)
                self.assertTrue("created" in item)
                # shouldn't have total_size, other verbose only,items
                self.assertFalse("total_size" in item)

            i += 1
        self.assertTrue(test_domain_found)
        self.assertEqual(i, count)
        d.close()

        # open in verbose mode
        d = h5py.Folder(folder_name, verbose=True)  # get folder object     

        self.assertEqual(d.domain, folder_name)
        self.assertTrue(d.modified)
        self.assertTrue(d.created)
        self.assertEqual(str(d), folder_name)
        self.assertEqual(d.owner, self.test_user1["name"])

        dir_parent = d.parent
        self.assertEqual(dir_parent[:-1], op.dirname(folder_name[:-1]))

        # get ACL for dir
        dir_acl = d.getACL(self.test_user1["name"])
        self.assertEqual(len(dir_acl.keys()), 7)
        for k in dir_acl.keys():
            self.assertTrue(dir_acl[k])

        dir_acls = d.getACLs()
        self.assertTrue(isinstance(dir_acls, list))

        count = len(d)
        self.assertTrue(count > 1)

        test_domain_found = False

        i = 0
        for name in d:
            if name == op.basename(test_domain):
                self.assertFalse(test_domain_found)
                test_domain_found = True
            item = d[name]
            #'owner': 'test_user1',
            #'created': 1496729517.2346532,
            #'class': 'domain',
            #'name': '/org/hdfgroup/h5pyd_test/bool_dset',
            #'lastModified': 1496729517.2346532
            #self.assertTrue("created" in item)
            self.assertTrue("owner" in item)
            self.assertTrue("class" in item)
            self.assertTrue("name" in item)
            if "root" in item:
                # non-folder objects will have last modified time
                self.assertTrue("lastModified" in item)
                self.assertTrue("created" in item)
                # these should show up only in verbose mode
                self.assertTrue("md5_sum" in item)
                self.assertTrue("num_groups" in item)
                self.assertTrue("num_datasets" in item)
                self.assertTrue("num_datatypes" in item)
                self.assertTrue("num_objects" in item)
                self.assertTrue("num_chunks" in item)
                self.assertTrue("num_linked_chunks" in item)
                self.assertTrue("total_size" in item)
                self.assertTrue("allocated_bytes" in item)
                self.assertTrue("metadata_bytes" in item)
                self.assertTrue("metadata_bytes" in item)
                self.assertTrue("linked_bytes" in item)
                self.assertTrue("compressors" in item)
            i += 1
        self.assertTrue(test_domain_found)
        self.assertEqual(i, count)
        d.close()


        # try opening a domain object as a folder
        f = h5py.Folder(filepath + '/')
        count = len(f)
        self.assertEqual(count, 0)
        for name in f:
            self.assertTrue(False)  # unexpected
        f.close()


    def test_create_folder(self):
        empty = self.getFileName("empty")
        empty_path = self.getPathFromDomain(empty)

        print("empty_path", empty_path)

        f = h5py.File(empty_path, mode='a')
        self.assertTrue(f.id.id is not None)
        if config.get("use_h5py"):
            # Folders not supported for h5py
            f.close()
            return

        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # HSDS currently supports folders, but h5serv does not
            f.close()
            return
        f.close()

        folder_test = self.getFileName("create_folder_test")
        folder_path = self.getPathFromDomain(folder_test) + '/'

        d = h5py.Folder(folder_path, mode='w')  # create a new folder
        d.close()
        # re-open
        d = h5py.Folder(folder_path)
        self.assertTrue(d.is_folder)
        d.close()


    def test_root_folder(self):
        test_domain = self.getFileName("folder_test")


        filepath = self.getPathFromDomain(test_domain)
        f = h5py.File(filepath, mode='a')
        self.assertTrue(f.id.id is not None)
        if config.get("use_h5py"):
            # Folders not supported for h5py
            f.close()
            return
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # HSDS currently supports folders, but h5serv does not
            f.close()
            return
        f.close()

        path_components = filepath.split('/')
        top_level_domain = path_components[1]

        d = h5py.Folder('/')  # get folder object for root
        found = False
        self.assertTrue(len(d) > 0)
        self.assertTrue(d.is_folder)
        self.assertTrue(d.domain == '/')
        self.assertTrue(d.__repr__() == '/')
        self.assertIsNone(d.parent)
        for name in d:
            # we should come across the given domain
            if top_level_domain == name:
                found = True
        d.close()
        self.assertTrue(found)

if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()




import sys
import os.path as op
import logging
from datetime import datetime
import h5pyd as h5py
import numpy as np
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Print objects in a domain in the style of the hsls utilitiy
#

cfg = Config()


def intToStr(n):
    if cfg["human_readable"]:
        s = "{:,}".format(n)
    else:
        s = "{}".format(n)
    return s


def format_size(n):
    if n is None or n == ' ':
        return ' ' * 8
    symbol = ' '
    if not cfg["human_readable"]:
        return str(n)
    # convert to common storage unit
    for s in ('B', 'K', 'M', 'G', 'T'):
        if n < 1024:
            symbol = s
            break
        n /= 1024
    if symbol == 'B':
        return "{:7}B".format(n)
    else:
        return "{:7.1f}{}".format(n, symbol)


def getShapeText(dset):
    shape_text = "{NULL}"
    shape = dset.shape
    if shape is not None:
        shape_text = "{"
        rank = len(shape)
        if rank == 0:
            shape_text += "{SCALAR}"
        else:
            for dim in range(rank):
                if dim != 0:
                    shape_text += ", "
                shape_text += str(shape[dim])
                if cfg["verbose"]:
                    # get unlimited dimension
                    max_extent = dset.maxshape[dim]
                    shape_text += '/'
                    if max_extent is None:
                        shape_text += "Inf"
                    else:
                        shape_text += str(max_extent)
        shape_text += "}"
    return shape_text


def visititems(name, grp, visited):
    for k in grp:
        item = grp.get(k, getlink=True)
        class_name = item.__class__.__name__
        item_name = op.join(name, k)
        if class_name == "HardLink":
            # follow hardlinks
            try:
                item = grp.get(k)
                dump(item_name, item, visited=visited)
            except IOError:
                # object deleted but hardlink left?
                desc = "{Missing hardlink object}"
                print("{0:24} {1} {2}".format(item_name, class_name, desc))

        elif class_name == "SoftLink":
            desc = '{' + item.path + '}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))
        elif class_name == "ExternalLink":
            desc = '{' + item.path + '//' + item.filename + '}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))
        else:
            desc = '{Unknown Link Type}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))


def getTypeStr(dt):
    # TBD: convert dt to more HDF5-like representation
    return str(dt)


def dump(name, obj, visited=None):
    class_name = obj.__class__.__name__
    desc = None
    obj_id = None
    if class_name in ("Dataset", "Group", "Datatype", "Table"):
        obj_id = obj.id.id
        if visited and obj_id in visited:
            same_as = visited[obj_id]
            print("{0:24} {1}, same as {2}".format(name, class_name, same_as))
            return
    elif class_name in ("ExternalLink", "SoftLink"):
        pass
    else:
        raise TypeError("unexpected classname: {}".format(class_name))

    is_dataset = False
    if class_name in ("Dataset", "Table"):
        is_dataset = True


    if is_dataset:
        desc = getShapeText(obj)
        obj_id = obj.id.id
    elif class_name == "Group":
        obj_id = obj.id.id
    elif class_name == "Datatype":
        obj_id = obj.id.id
    elif class_name == "SoftLink":
        desc = '{' + obj.path + '}'
    elif class_name == "ExternalLink":
        desc = '{' + obj.filename + '//' + obj.path + '}'

    if desc is None:
        print("{0} {1}".format(name, class_name))
    else:
        print("{0} {1} {2}".format(name, class_name, desc))

    if cfg["verbose"] and obj_id is not None:
        print("    {0:>32}: {1}".format("UUID", obj_id))

    if cfg["verbose"] and is_dataset and obj.shape is not None \
            and obj.chunks is not None:
        chunk_size = obj.dtype.itemsize
        if isinstance(obj.chunks, dict):
            # H5D_CHUNKED_REF layout
            chunk_dims = obj.chunks["dims"]
            storage_desc = "Storage " + obj.chunks["class"]
        else:
            chunk_dims = obj.chunks
            storage_desc = "Storage H5D_CHUNKED"
        for chunk_dim in chunk_dims:
            chunk_size *= chunk_dim
        dset_size = obj.dtype.itemsize
        for dim_extent in obj.shape:
            dset_size *= dim_extent

        num_chunks = obj.num_chunks
        allocated_size = obj.allocated_size
        if num_chunks is not None and allocated_size is not None:
            fstr = "    {0:>32}: {1} {2} bytes, {3} allocated chunks"
            print(fstr.format("Chunks", chunk_dims, intToStr(chunk_size),
                              intToStr(num_chunks)))
            if dset_size > 0:
                utilization = allocated_size / dset_size
                fstr = "    {0:>32}: {1} logical bytes, {2} allocated bytes, {3:.2f}% utilization"
                print(fstr.format(storage_desc, intToStr(dset_size),
                                  intToStr(allocated_size),
                                  utilization * 100.0))
            else:
                fstr = "    {0:>32}: {1} logical bytes, {2} allocated bytes"
                print(fstr.format(storage_desc, intToStr(dset_size),
                                  intToStr(allocated_size)))

        else:
            # verbose info not available, just show the chunk layout
            fstr = "    {0:>32}: {1} {2} bytes"
            print(fstr.format("Chunks", chunk_dims, intToStr(chunk_size)))

        fstr = "    {0:>32}: {1}"
        print(fstr.format("Type", getTypeStr(obj.dtype)))  # dump out type info

    if cfg["showattrs"] and class_name in ("Dataset", "Table", "Group", "Datatype"):
        # dump attributes for the object
        for attr_name in obj.attrs:
            attr = obj.attrs[attr_name]
            el = "..."  # show this if the attribute is too large
            if isinstance(attr, np.ndarray):
                rank = len(attr.shape)
            else:
                rank = 0  # scalar data
            if rank > 1:
                val = "[" * rank + el + "]" * rank
                print("   attr: {0:24} {1}".format(attr_name, val))
            elif rank == 1 and attr.shape[0] > 1:
                val = "[{},{}]".format(attr[0], el)
                print("   attr: {0:24} {1}".format(attr_name, val))
            else:
                print("   attr: {0:24} {1}".format(attr_name, attr))

    if visited is not None and obj_id is not None:
        visited[obj_id] = name
    if class_name == "Group" and visited is not None:
        visititems(name, obj, visited)


def dumpACL(acl):
    perms = ""
    if acl["create"]:
        perms += 'c'
    else:
        perms += '-'
    if acl["read"]:
        perms += 'r'
    else:
        perms += '-'
    if acl["update"]:
        perms += 'u'
    else:
        perms += '-'
    if acl["delete"]:
        perms += 'd'
    else:
        perms += '-'
    if acl["readACL"]:
        perms += 'e'
    else:
        perms += '-'
    if acl["updateACL"]:
        perms += 'p'
    else:
        perms += '-'
    print("    acl: {0:24} {1}".format(acl["userName"], perms))


def dumpAcls(obj):
    try:
        acls = obj.getACLs()
    except IOError:
        print("read ACLs is not permitted")
        return

    for acl in acls:
        dumpACL(acl)


def getFolder(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket   = cfg["hs_bucket"]
    pattern = cfg["pattern"]
    query = cfg["query"]
    if cfg["verbose"]:
        verbose = True
    else:
        verbose = False
    batch_size = 100  # use smaller batchsize for interactively listing of large collections
    d = h5py.Folder(domain, endpoint=endpoint, username=username, verbose=verbose,
                      password=password, bucket=bucket, pattern=pattern, query=query, batch_size=batch_size)
    return d


def getFile(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    fh = h5py.File(domain, mode='r', endpoint=endpoint, username=username,
                   password=password, bucket=bucket, use_cache=True)
    return fh


def visitDomains(domain, depth=1):
    if depth == 0:
        return 0

    count = 0
    if domain[-1] == '/':
        domain = domain[:-1]  # strip off trailing slash

    try:
        d = getFolder(domain + '/')
        dir_class = "domain"
        display_name = domain
        num_bytes = ' '
        if d.is_folder:
            dir_class = "folder"
            display_name += '/'
        elif cfg["verbose"]:
            # get the number of allocated bytes
            f = getFile(domain)
            num_bytes = f.total_size
            f.close()

        owner = d.owner
        if owner is None:
            owner = ""
        if d.modified is None:
            timestamp = ""
        else:
            timestamp = datetime.fromtimestamp(int(d.modified))

        print("{:35} {:15} {:8} {} {}".format(owner, format_size(num_bytes),
                                              dir_class, timestamp,
                                              display_name))
        count += 1
        if cfg["showacls"]:
            dumpAcls(d)
        for name in d:
            item = d[name]
            owner = item["owner"]
            full_path = domain + '/' + name

            num_bytes = " "
            if cfg["verbose"] and "total_size" in item:
                num_bytes = item["total_size"]
            else:
                 num_bytes = " "
            dir_class = item["class"]
            if item["lastModified"] is None:
                timestamp = ""
            else:
                timestamp = datetime.fromtimestamp(int(item["lastModified"]))

            print("{:35} {:15} {:8} {} {}".format(owner, format_size(num_bytes),
                                              dir_class, timestamp,
                                              full_path))
            count += 1

            if dir_class == "folder":
                # recurse for folders
                n = visitDomains(domain + '/' + name, depth=(depth - 1))
                count += n

    except IOError as oe:
        if oe.errno in (403, 404, 410):
            # TBD: recently created domains may not be immediately visible to
            # the service Once the flush operation is implemented, this should
            # be an issue for h5pyd apps
            #
            # Also, ignore domains for which we don't have permsssions (403)
            pass
        else:
            print("error getting domain:", domain)
            sys.exit(str(oe))

    return count



#
# Usage
#
def printUsage():
    print("usage: {} [-v] [-h] [--showacls] [--showattrs] [--recursive|-r] [--loglevel debug|info|warning|error] [--logfile <logfile>] [-e endpoint] [-u username] [-p password] [--bucket bucketname] domains".format(cfg["cmd"]))
    print("example: {} -r -e http://hsdshdflab.hdfgroup.org /shared/tall.h5".format(cfg["cmd"]))
    print("")
    print("Options:")
    print("     -v | --verbose :: verbose output")
    print("     -H | --human-readable :: with -v, print human readable sizes (e.g. 123M)")
    print("     -e | --endpoint <domain> :: The HDF Server endpoint, e.g. http://hsdshdflab.hdfgroup.org")
    print("     -u | --user <username>   :: User name credential")
    print("     -p | --password <password> :: Password credential")
    print("     -c | --conf <file.cnf>  :: A credential and config file")
    print("     --showacls :: prints domain ACLs")
    print("     --showattrs :: print attributes")
    print("     --pattern  :: <regex>  :: list domains that match the given regex")
    print("     --query :: <query> list domains where the attributes of the root group match the given query string")
    print("     --logfile <logfile> :: logfile path")
    print("     --loglevel debug|info|warning|error :: Change log level")
    print("     --bucket <bucket_name> :: Storage bucket")
    print("     --recursive, -r :: recursively list sub-folders or sub-groups")
    print("     -h | --help    :: This message.")
    sys.exit()


#
# Main
#
def main():
    domains = []
    argn = 1
    depth = 1
    loglevel = logging.ERROR
    logfname = None
    cfg["verbose"] = False
    cfg["showacls"] = False
    cfg["showattrs"] = False
    cfg["human_readable"] = False
    cfg["pattern"] = None
    cfg["query"] = None
    cfg["cmd"] = sys.argv[0].split('/')[-1]
    if cfg["cmd"].endswith(".py"):
        cfg["cmd"] = "python " + cfg["cmd"]

    while argn < len(sys.argv):
        arg = sys.argv[argn]
        val = None
        if len(sys.argv) > argn + 1:
            val = sys.argv[argn + 1]
        if arg in ("-r", "--recursive"):
            depth = -1
            argn += 1
        elif arg in ("-v", "--verbose"):
            cfg["verbose"] = True
            argn += 1
        elif arg in ("-H", "--human-readable"):
            cfg["human_readable"] = True
            argn += 1
        elif arg == "--loglevel":
            val = val.upper()
            if val == "DEBUG":
                loglevel = logging.DEBUG
            elif val == "INFO":
                loglevel = logging.INFO
            elif val in ("WARN", "WARNING"):
                loglevel = logging.WARNING
            elif val == "ERROR":
                loglevel = logging.ERROR
            else:
                printUsage()
            argn += 2
        elif arg == '--logfile':
            logfname = val
            argn += 2
        elif arg in ("-showacls", "--showacls"):
            cfg["showacls"] = True
            argn += 1
        elif arg in ("-showattrs", "--showattrs"):
            cfg["showattrs"] = True
            argn += 1
        elif arg in ("-h", "--help"):
            printUsage()
        elif arg in ("-e", "--endpoint"):
            cfg["hs_endpoint"] = val
            argn += 2
        elif arg in ("-u", "--username"):
            cfg["hs_username"] = val
            argn += 2
        elif arg in ("-p", "--password"):
            cfg["hs_password"] = val
            argn += 2
        elif arg in ("-b", "--bucket"):
            cfg["hs_bucket"] = val
            argn += 2
        elif arg == "--pattern":
            cfg["pattern"] = val
            argn += 2
        elif arg == "--query":
            cfg["query"] = val
            argn += 2

        elif arg[0] == '-':
            printUsage()
        else:
            domains.append(arg)
            argn += 1

    # setup logging
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s',
                        level=loglevel)
    logging.debug("set log_level to {}".format(loglevel))

    if len(domains) == 0:
        # add top-level domain
        domains.append("/")

    for domain in domains:
        if domain.endswith('/'):
            # given a folder path
            count = visitDomains(domain, depth=depth)
            print("{} items".format(count))

        else:
            try:
                f = getFile(domain)
            except IOError as ioe:
                if ioe.errno == 401:
                    print("Username/Password missing or invalid")
                    continue
                if ioe.errno == 403:
                    print("No permission to read domain: {}".format(domain))
                    continue
                elif ioe.errno == 404:
                    print("Domain {} not found".format(domain))
                    continue
                elif ioe.errno == 410:
                    print("Domain {} has been removed".format(domain))
                    continue
                else:
                    print("Unexpected error: {}".format(ioe))
                    continue

            grp = f['/']
            if grp is None:
                print("{}: No such domain".format(domain))
                domain += '/'
                count = visitDomains(domain, depth=depth)
                print("{} items".format(count))
                continue
            dump('/', grp)

            if depth < 0:
                # recursive
                visited = {}  # dict of id to h5path
                visited[grp.id.id] = '/'
                visititems('/', grp, visited)
            else:
                for k in grp:
                    item = grp.get(k, getlink=True)
                    if item.__class__.__name__ == "HardLink":
                        # follow hardlinks
                        try:
                            item = grp[k]
                        except IOError:
                            # object deleted?  Just dump link info
                            pass
                    dump(k, item)
            if cfg["showacls"]:
                dumpAcls(grp)
            grp.file.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including       s   #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import os
import sys

hl_tests = ('test_attribute',
            'test_committedtype',
            'test_complex_numbers',
            'test_dataset_compound',
            'test_dataset_create',
            'test_dataset_extend',
            'test_dataset_fancyselect',
            'test_dataset_objref',
            'test_dataset_getitem',
            'test_dataset_pointselect',
            'test_dataset_scalar',
            'test_dataset_setitem',
            'test_dimscale',
            'test_file',
            'test_folder',
            'test_group',
            'test_table',
            'test_visit',
            'test_vlentype')

app_tests = ('test_hsinfo', 'test_tall_inspect', 'test_diamond_inspect',
             'test_shuffle_inspect')

run_hl = True
run_app = True
if len(sys.argv) > 1:
    if sys.argv[1] in ("-h", "--help"):
        print("Usage: python testall.py [--hl_only | --app_only]")
        sys.exit(0)
    if sys.argv[1] == "--hl_only":
        run_app = False
    elif sys.argv[1] == "--app_only":
        run_hl = False
    else:
        print("Unexpected argument")
        sys.exit(1)

#
# Run tests
#
os.chdir('test')
if run_hl:
    os.chdir('hl')
    for test_name in hl_tests:
        print(test_name)
        rc = os.system('python ' + test_name + '.py')
        if rc != 0:
            sys.exit("Failed")
    os.chdir('..')
if run_app:
    os.chdir('apps')
    rc = os.system('python is_hsds.py')
    if rc == 0:
        print("running HSDS app tests")
        # these tests are only support with HSDS
        rc = os.system('python load_files.py')
        if rc != 0:
            sys.exit("load_files.py failed")
        for test_name in app_tests:
            print(test_name)
            rc = os.system('python ' + test_name + '.py')
            if rc != 0:
                sys.exit("Failed")

os.chdir('..')

print("Done!")



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################


import numpy as np

CHUNK_BASE = 16*1024    # Multiplier by which chunks are adjusted
CHUNK_MIN = 8*1024      # Soft lower limit (8k)
CHUNK_MAX = 1024*1024   # Hard upper limit (1M)

def guess_chunk(shape, maxshape, typesize):
    """ Guess an appropriate chunk layout for a dataset, given its shape and
    the size of each element in bytes.  Will allocate chunks only as large
    as MAX_SIZE.  Chunks are generally close to some power-of-2 fraction of
    each axis, slightly favoring bigger values for the last index.

    Undocumented and subject to change without warning.
    """
    # pylint: disable=unused-argument

    # For unlimited dimensions we have to guess 1024
    shape = tuple((x if x!=0 else 1024) for i, x in enumerate(shape))

    ndims = len(shape)
    if ndims == 0:
        raise ValueError("Chunks not allowed for scalar datasets.")

    chunks = np.array(shape, dtype='=f8')
    if not np.all(np.isfinite(chunks)):
        raise ValueError("Illegal value in chunk tuple")

    # Determine the optimal chunk size in bytes using a PyTables expression.
    # This is kept as a float.
    dset_size = np.product(chunks)*typesize
    target_size = CHUNK_BASE * (2**np.log10(dset_size/(1024.*1024)))

    if target_size > CHUNK_MAX:
        target_size = CHUNK_MAX
    elif target_size < CHUNK_MIN:
        target_size = CHUNK_MIN

    idx = 0
    while True:
        # Repeatedly loop over the axes, dividing them by 2.  Stop when:
        # 1a. We're smaller than the target chunk size, OR
        # 1b. We're within 50% of the target chunk size, AND
        #  2. The chunk is smaller than the maximum chunk size

        chunk_bytes = np.product(chunks)*typesize

        if (chunk_bytes < target_size or \
         abs(chunk_bytes-target_size)/target_size < 0.5) and \
         chunk_bytes < CHUNK_MAX:
            break

        if np.product(chunks) == 1:
            break  # Element size larger than CHUNK_MAX

        chunks[idx%ndims] = np.ceil(chunks[idx%ndims] / 2.0)
        idx += 1

    return tuple(int(x) for x in chunks)
    
class ChunkIterator:
    """
    Class to iterate through list of chunks given h5py dset
    """
    def __init__(self, dset):
        self._shape = dset.shape
        if dset.chunks is None:
             # guess a chunk shape
            self._layout =  guess_chunk(self._shape, None, dset.dtype.itemsize)
            print(f"guess chunk: {self._layout}")
        elif isinstance(dset.chunks, dict):
            self._layout = dset.chunks["dims"]
        else:
            self._layout = dset.chunks
        self._rank = len(dset.shape)
        if self._rank == 0:
            self._chunk_index = [0,]
        else:
            self._chunk_index = [0,] * self._rank

    def __iter__(self):
        return self

    def __next__(self):

        def get_ret(item):
            if len(item) == 1:
                return item[0]
            else:
                return tuple(item)
        if self._layout == ():
            # special case for scalar datasets
            if self._chunk_index[0] > 0:
                raise StopIteration()
            self._chunk_index[0] += 1
            return ()

        slices = []
        if self._chunk_index[0] * self._layout[0] >= self._shape[0]:
            # ran past the last chunk, end iteration
            raise StopIteration()

        for dim in range(self._rank):
            start = self._chunk_index[dim] * self._layout[dim]
            stop = start + self._layout[dim]
            if stop > self._shape[dim]:
                stop = self._shape[dim]  # trim to end of dimension
            s = slice(start, stop, 1)
            slices.append(s)

        # bump up the last index and carry forward if we run outside the selection
        dim = self._rank - 1
        while dim >= 0:
            c = self._layout[dim]
            self._chunk_index[dim] += 1

            chunk_end = self._chunk_index[dim] * c
            if chunk_end < self._shape[dim]:
                # we still have room to extend along this dimensions
                return get_ret(slices)

            if dim > 0:
                # reset to the start and continue iterating with higher dimension
                self._chunk_index[dim] = 0
            dim -= 1
        return get_ret(slices)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import os

def get_test_filenames():
    return ("tall.h5", "diamond.h5", "shuffle_compress.h5")

cfg = {
    'use_h5py': False
}

def get(x):
    # see if there is a command-line override
    config_value = None

    # see if there are an environment variable override
    if x.upper() in os.environ:
        config_value = os.environ[x.upper()]
    # no command line override, just return the cfg value
    if config_value is None:
        config_value = cfg[x]

    # convert string to boolean if true or false
    if type(config_value) is str:
        if config_value.upper() in ('T', 'TRUE'):
            config_value = True
        elif config_value.upper() in ('F', 'FALSE'):
            config_value = False
    return config_value






##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import os
import config
import h5pyd

if config.get("use_h5py"):
    sys.exit("use_h5py")

if "H5PYD_TEST_FOLDER" not in os.environ:
    sys.exit("set H5PYD_TEST_FOLDER environment not set")
folder_path = os.environ["H5PYD_TEST_FOLDER"]
if not folder_path.startswith("/"):
    # HSDS expects folder paths to start with a slash (as opposed to DNS format)
    sys.exit("not HSDS path")
try:
    h5pyd.Folder(folder_path)  # will trigger error with h5serv
except Exception:
    sys.exit("Server doesn't support Folder objects")
    







##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import os
import numpy as np
import config
import logging
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase

def get_filename():
    if config.get("use_h5py"):
        dirpath = "data" 
    else:
        dirpath = config.get("H5PYD_TEST_FOLDER")
    filename = os.path.join(dirpath, "tall.h5")
    return filename


class TestTallInspect(TestCase):

    def setUp(self):
        
        filename = get_filename()
        self.f = h5py.File(filename, "r")

    def test_obj_count(self):
        counts = {"groups": 0, "datasets": 0}
        def visit(name):
            obj = self.f[name]
            if isinstance(obj, h5py.Dataset):
                counts["datasets"] += 1
            elif isinstance(obj, h5py.Group):
                counts["groups"] += 1
            else:
                # unexpected class
                self.assertTrue(False)
        self.f.visit(visit)
        self.assertEqual(5, counts["groups"])
        self.assertEqual(4, counts["datasets"])

    def test_attributes(self):
        self.assertEqual(2, len(self.f.attrs))
        self.assertTrue("attr1" in self.f.attrs)
        attr1 = self.f.attrs["attr1"]
        self.assertEqual(attr1.dtype, np.int8)
        self.assertEqual(attr1.shape, (10,))
        for i in range(10):
            self.assertEqual(attr1[i], ((97 + i) % 106))
        self.assertTrue("attr2" in self.f.attrs)
        attr2 = self.f.attrs["attr2"]
        self.assertEqual(attr2.dtype, np.dtype(">i4"))
        self.assertEqual(attr2.shape, (2,2))
        for i in range(2):
            for j in range(2):
                self.assertEqual(attr2[i,j], 2*i+j)
        dset111 = self.f["/g1/g1.1/dset1.1.1"]
        self.assertEqual(2, len(dset111.attrs))
        self.assertTrue("attr1" in dset111.attrs)
        self.assertTrue("attr2" in dset111.attrs)

    def test_links(self):
        g12 = self.f["/g1/g1.2"]
        self.assertEqual(len(g12), 2)
        self.assertTrue("extlink" in g12)
        extlink = g12.get("extlink", getlink=True)
        self.assertEqual(extlink.__class__.__name__, "ExternalLink")
        self.assertEqual(extlink.filename, "somefile")
        self.assertEqual(extlink.path, "somepath")
        g121 = self.f["/g1/g1.2/g1.2.1"]
        self.assertEqual(len(g121), 1)
        softlink = g121.get("slink", getlink=True)
        self.assertEqual(softlink.path, "somevalue")

    def test_dataset(self):
        dset112 = self.f["/g1/g1.1/dset1.1.2"]
        self.assertEqual(dset112.shape, (20,))
        self.assertEqual(dset112.dtype, np.dtype(">i4"))
        arr = dset112[...]
        for i in range(20):
            self.assertEqual(arr[i], i)

if __name__ == '__main__':
    print("filename:", get_filename())
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    
    ut.main()




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import logging
import config
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase


class TestObjRef(TestCase):


    def test_create(self):
        filename = self.getFileName("objref_test")
        print(filename)

        f = h5py.File(filename, 'w')
        self.assertTrue(f.id.id is not None)
        self.assertTrue('/' in f)
        r = f['/']
        is_h5serv = False
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            is_h5serv = True  # h5serv doesn't have support for objref datasets yet

        # create subgroup g1
        r.create_group('g1')
        self.assertTrue('g1' in r)
        g1 = r['g1']

        # create subgroup g1/g1.1
        g11 = g1.create_group('g1.1')

        # get ref to g1/g1.1
        g11_ref = g11.ref
        self.assertTrue(isinstance(g11_ref, h5py.Reference))
        self.assertEqual(g11.name, "/g1/g1.1")

        # create subgroup g2
        r.create_group('g2')
        self.assertEqual(len(r), 2)
        g2 = r['g2']

        # get ref to g1/g1.1 from g2
        g11ref = g2[g11_ref]

        # create subgroup /g1/g1.1/foo
        g11ref.create_group("foo")
        self.assertEqual(len(g11), 1)
        self.assertTrue("foo" in g11)

        # create datset /g2/d1
        d1 = g2.create_dataset('d1', (10,), dtype='i8')

        # get ref to d1
        d1_ref = d1.ref
        dt = h5py.special_dtype(ref=h5py.Reference)
        self.assertTrue(dt.metadata['ref'] is h5py.Reference)
        ref = h5py.check_dtype(ref=dt)
        self.assertEqual(ref, h5py.Reference)


        if is_h5serv:
            return  # ref types not supported in h5serv

        # create dataset of ref types
        dset = g1.create_dataset('myrefs', (10,), dtype=dt)
        ref = h5py.check_dtype(ref=dset.dtype)
        self.assertEqual(ref, h5py.Reference)

        dset[0] = g11_ref
        dset[1] = d1_ref

        a_ref = dset[0]
        obj = f[a_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, g11.id.id)  # ref to g1.1
        self.assertEqual(obj.name, "/g1/g1.1")

        b_ref = dset[1]
        obj = f[b_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, d1.id.id)  # ref to d1
        self.assertEqual(obj.name, "/g2/d1")

        # try the same thing using attributes
        ref_values = [g11_ref, d1_ref]
        g1.attrs.create("a1", ref_values, dtype=dt)

        # pass a single ref to attribute create
        obj = f[g11_ref]
        g1.attrs.create("b1", g11_ref, dtype=dt)

        # read back the attribute
        attr = g1.attrs["a1"]
        self.assertEqual(attr.shape, (2,))
        ref = h5py.check_dtype(ref=attr.dtype)
        self.assertEqual(ref, h5py.Reference)
        a0_ref = attr[0]
        obj = f[a0_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, g11.id.id)  # ref to g1.1
        self.assertEqual(obj.name, "/g1/g1.1")
        a1_ref = attr[1]
        obj = f[a1_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, d1.id.id)  # ref to d1
        self.assertEqual(obj.name, "/g2/d1")
        f.close()


        # try opening in read-mode
        f = h5py.File(filename, 'r')
        g1 = f['/g1']
        g11 = f['g1/g1.1']
        d1 = f["/g2/d1"]
        attr = g1.attrs["a1"]
        self.assertEqual(attr.shape, (2,))
        ref = h5py.check_dtype(ref=attr.dtype)
        self.assertEqual(ref, h5py.Reference)
        a0_ref = attr[0]
        obj = f[a0_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, g11.id.id)  # ref to g1.1

        self.assertEqual(obj.name, "/g1/g1.1")
        a1_ref = attr[1]
        obj = f[a1_ref]
        if not config.get("use_h5py"):
            self.assertEqual(obj.id.id, d1.id.id)  # ref to d1
        self.assertEqual(obj.name, "/g2/d1")



    def test_delete(self):
        filename = self.getFileName("objref_delete_test")
        print(filename)
        f = h5py.File(filename, 'w')
        self.assertTrue(f.id.id is not None)
        self.assertTrue('/' in f)
        is_h5serv = False
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            is_h5serv = True  # h5serv doesn't have support for objref datasets yet

        # create a dataset
        dset = f.create_dataset('dset', data=[1,2,3])
        dset_ref = dset.ref

        if not is_h5serv:
            f.attrs["dset_ref"] = dset_ref
            del f['dset']
            try:
                obj = f[dset_ref]
                if config.get("use_h5py"):
                    # TBD - HSDS is not triggering this exception since the object
                    # is not being deleted.
                    self.assertTrue(False)
            except ValueError:
                pass # expected

        f.close()


if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import numpy as np
import logging
import math
import os
import config
from common import ut, TestCase

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py


@ut.skipIf(os.environ.get('H5_DRIVER', '') == 'h5serv',
           'Does not work with h5serv due to an h5json exception')
class TestComplexNumbers(TestCase):
    """ Test read/write of complex numbers """

    def test_complex_dset(self):
        """Read and write complex numbers in a dataset"""
        filename = self.getFileName('create_complex_dset')
        print('filename:', filename)
        count = 10
        dt = np.dtype('complex64')
        with h5py.File(filename, 'w') as f:
            dset = f.create_dataset('complex', shape=(count,), dtype=dt)
            if not config.get('use_h5py'):
                # only h5pyd is setting up field names?
                self.assertEqual(dset.dtype.names, ('r', 'i'))
            for i in range(count):
                theta = (4.0 * math.pi) * (float(i) / float(count))
                dset[i] = math.cos(theta) + 1j * math.sin(theta)
            val = dset[0]

        self.assertEqual(val.shape, ())
        self.assertEqual(val.dtype.kind, 'c')
        self.assertEqual(val.real, 1.0)
        self.assertEqual(val.imag, 0.)

    def test_complex_attr(self):
        """Read and wrtie complex numbers in attributes"""
        filename = self.getFileName('create_complex_dset')
        print('filename:', filename)
        with h5py.File(filename, 'w') as f:
            dset = f.create_dataset('x', data=5)
            dset.attrs['scalar'] = 4 + 3 * 1j
            dset.attrs['array'] = [1 + 2 * 1j, 3 + 4 * 1j]

            scalar = dset.attrs['scalar']
            array = dset.attrs['array']

        self.assertEqual(scalar.shape, ())
        self.assertEqual(array.shape, (2,))
        self.assertEqual(scalar.dtype.kind, 'c')
        self.assertEqual(array.dtype.kind, 'c')
        self.assertEqual(scalar.real, 4.0)
        self.assertEqual(scalar.imag, 3.)
        self.assertEqual(array.real.tolist(), [1., 3.])
        self.assertEqual(array.imag.tolist(), [2., 4.])


if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import numpy as np
import math
import logging

import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase


class TestTable(TestCase):
    def test_create_table(self):
        filename = self.getFileName("create_table_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            return # Table not supported with h5py
        f = h5py.File(filename, "w")
        if not f.id.id.startswith("g-"):
            return # append not supported with h5serv

        count = 10

        dt = np.dtype([('real', float), ('img', float)])
        table = f.create_table('complex', numrows=10, dtype=dt)

        elem = table[0]
        for i in range(count):
            theta = (4.0 * math.pi)*(float(i)/float(count))
            elem['real'] = math.cos(theta)
            elem['img'] = math.sin(theta)
            table[i] = elem

        self.assertEqual(table.colnames, ['real', 'img'])
        self.assertEqual(table.nrows, count)

        num_rows = 0
        for row in table:
            self.assertEqual(len(row), 2)
            num_rows += 1
        self.assertEqual(num_rows, count)

        # try the same thing using cursor object
        cursor = table.create_cursor()
        num_rows = 0
        for row in cursor:
            self.assertEqual(len(row), 2)
            num_rows += 1
        self.assertEqual(num_rows, count)

        arr = table.read(start=5, stop=6)
        self.assertEqual(arr.shape, (1,))


        f.close()

    def test_query_table(self):
        filename = self.getFileName("query_compound_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            return # Table not supported with h5py
        f = h5py.File(filename, "w")

        if not f.id.id.startswith("g-"):
            return # append not supported with h5serv

        # write entire array
        data = [
            ("EBAY", "20170102", 3023, 3088),
            ("AAPL", "20170102", 3054, 2933),
            ("AMZN", "20170102", 2973, 3011),
            ("EBAY", "20170103", 3042, 3128),
            ("AAPL", "20170103", 3182, 3034),
            ("AMZN", "20170103", 3021, 2788),
            ("EBAY", "20170104", 2798, 2876),
            ("AAPL", "20170104", 2834, 2867),
            ("AMZN", "20170104", 2891, 2978),
            ("EBAY", "20170105", 2973, 2962),
            ("AAPL", "20170105", 2934, 3010),
            ("AMZN", "20170105", 3018, 3086)
        ]

        dt = np.dtype([('symbol', 'S4'), ('date', 'S8'), ('open', 'i4'), ('close', 'i4')])
        table = f.create_table('stock', dtype=dt)

        table.append(data)

        self.assertEqual(table.nrows, len(data))

        for indx in range(len(data)):
            row = table[indx]
            item = data[indx]
            for col in range(2,3):
                # first two columns will come back as bytes, not strs
                self.assertEqual(row[col], item[col])

        cursor = table.create_cursor()
        indx = 0
        for row in cursor:
            item = data[indx]
            for col in range(2,3):
                # first two columns will come back as bytes, not strs
                self.assertEqual(row[col], item[col])
            indx += 1
        self.assertEqual(indx, len(data))

        cursor = table.create_cursor(start=2, stop=5)
        indx = 2
        for row in cursor:
            item = data[indx]
            for col in range(2,3):
                # first two columns will come back as bytes, not strs
                self.assertEqual(row[col], item[col])
            indx += 1
        self.assertEqual(indx, 5)

        condition = "symbol == b'AAPL'"
        quotes = table.read_where(condition)
        self.assertEqual(len(quotes), 4)
        for i in range(4):
            quote = quotes[i]
            self.assertEqual(quote[0], b'AAPL')

        # read up to 2 rows
        quotes = table.read_where(condition, limit=2)
        self.assertEqual(len(quotes), 2)

        # use a query cursor
        cursor = table.create_cursor(condition=condition)
        num_rows = 0
        for row in cursor:
            self.assertEqual(len(row), 4)
            num_rows += 1
        self.assertEqual(num_rows, 4)
 
        # try a compound query
        condition = "(open > 3000) & (open < 3100)" 
        quotes = table.read_where(condition)
        self.assertEqual(len(quotes), 5)
        for i in range(4):
            quote = quotes[i]
            self.assertTrue(quote[2] > 3000)
            self.assertTrue(quote[2] < 3100)
        
        # try modifying specific rows
        condition = "symbol == b'AAPL'"
        update_val = {"open": 123}
        indices = table.update_where(condition, update_val)
        self.assertEqual(len(indices), 4)
        self.assertEqual(list(indices), [1,4,7,10])

        row = tuple(table[4])
        self.assertEqual(row, (b'AAPL', b'20170103', 123, 3034))

        # try modifying just one value
        update_val = {'close': 999}
        indices = table.update_where(condition, update_val, limit=1)
        self.assertEqual(len(indices), 1)
        self.assertEqual(list(indices), [1])
        f.close()

if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()
